import os
import logging
import json
from datetime import datetime, timedelta
import pandas as pd
from dotenv import load_dotenv
import time
import requests

# Setup Logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

# Load environment variables from .env file located in the config directory
CONFIG_DIR = 'C:\\Users\\user\\projects\\st\\config'
load_dotenv(dotenv_path=os.path.join(CONFIG_DIR, '.env'))

# Constants
RETRIES = 3
DELAY = 60  # Delay between retries
DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
SHORT_DATE_FORMAT = "%Y%m%d"
MAX_ATTEMPTS = 10
PAGINATION_DELAY = 0.05  # Delay between pagination API requests

# Load settings
def load_settings():
    settings = {
        'data_path': os.getenv('DATA_PATH'),
        'model_path': os.getenv('MODEL_PATH'),
        'approval_key': os.getenv('APPROVAL_KEY'),
        'max_detected_stocks': int(os.getenv('MAX_DETECTED_STOCKS', 3000)),
        'data_fetch_interval': int(os.getenv('DATA_FETCH_INTERVAL', 60)),
        'strategy_name': os.getenv('STRATEGY_NAME'),
        'token_path': os.getenv('TOKEN_PATH'),
        'appkey': os.getenv('APPKEY'),
        'appsecret': os.getenv('APPSECRET'),
        'user_id': os.getenv('USER_ID'),
        'trading_url': os.getenv('TRADING_URL'),
        'agent': os.getenv('AGENT'),
        'acct_stock': os.getenv('ACCT_STOCK'),
        'prod': os.getenv('PROD'),
        'cano': os.getenv('CANO')
    }

    config_json_path = os.path.join(CONFIG_DIR, 'config.json')
    if os.path.exists(config_json_path):
        with open(config_json_path, 'r') as f:
            config_settings = json.load(f)
            settings.update(config_settings)
    else:
        raise FileNotFoundError(f"{config_json_path} not found.")
    
    return settings

settings = load_settings()

# Token Handling Functions
def get_token_data():
    """Retrieve token data from the token file."""
    token_path = settings['token_path']
    try:
        with open(token_path, 'r') as file:
            return json.load(file)
    except (FileNotFoundError, json.JSONDecodeError) as e:
        logging.error(f"Error reading token file: {e}")
        return None

def save_token_data(token_data):
    """Save token data to the token file."""
    token_path = settings['token_path']
    os.makedirs(os.path.dirname(token_path), exist_ok=True)
    with open(token_path, 'w') as file:
        json.dump(token_data, file)

def is_token_valid(token_data):
    """Check if the token is valid based on its expiration time."""
    if not token_data or 'access_token' not in token_data:
        return False
    expiration_str = token_data.get('expiration')
    if not expiration_str:
        return False
    expiration = datetime.strptime(expiration_str, DATE_FORMAT)
    return expiration > datetime.now()

def get_valid_token():
    """Retrieve a valid token, refreshing it if necessary."""
    token_data = get_token_data()
    if is_token_valid(token_data):
        logging.info("Using existing token")
        return token_data['access_token']
    else:
        logging.info("Token expired or invalid, refreshing...")
        return refresh_token()

def refresh_token(retries=RETRIES, delay=DELAY):
    """Refresh the access token, retrying on failure."""
    url = f"{settings['trading_url']}/oauth2/tokenP"
    data = {
        'grant_type': 'client_credentials',
        'appkey': settings['appkey'],
        'appsecret': settings['appsecret']
    }
    headers = {'Content-Type': 'application/json'}
    
    for attempt in range(retries):
        try:
            logging.info("Requesting new token...")
            response = requests.post(url, headers=headers, json=data)
            if response.status_code == 200:
                response_data = response.json()
                new_token = response_data.get('access_token')
                expires_in = response_data.get('access_token_token_expired')
                expiration = datetime.now() + timedelta(seconds=int(expires_in) if expires_in else 3600)
                token_data = {'access_token': new_token, 'expiration': expiration.strftime(DATE_FORMAT)}
                save_token_data(token_data)
                logging.info(f"New token saved. Expiration time: {expiration}")
                return new_token
            else:
                logging.error(f"Error refreshing token: {response.text}")
                if "EGW00133" in response.text:
                    logging.info(f"Retrying in {delay} seconds...")
                    time.sleep(delay)
                else:
                    return None
        except requests.exceptions.RequestException as e:
            logging.error(f"Request exception occurred while refreshing token: {e}")
            return None
    return None

# API Request Functions
def get_api_response(url, headers, params):
    """Send a GET request to the specified URL with headers and parameters."""
    try:
        response = requests.get(url, headers=headers, params=params)
        response.raise_for_status()
        return response.json()
    except requests.exceptions.RequestException as e:
        logging.error(f"Request exception occurred: {e}")
        return None

def fetch_stock_data(stock_code, start_date, end_date):
    """Fetch stock data for a given stock code and date range."""
    url = f"{settings['trading_url']}/uapi/domestic-stock/v1/quotations/inquire-daily-itemchartprice"
    headers = {
        'Content-Type': 'application/json; charset=utf-8',
        'Authorization': f'Bearer {get_valid_token()}',
        'appkey': settings['appkey'],
        'appsecret': settings['appsecret'],
        'tr_id': 'FHKST03010100'
    }
    params = {
        'FID_COND_MRKT_DIV_CODE': 'J',
        'FID_INPUT_ISCD': stock_code,
        'FID_INPUT_DATE_1': start_date.strftime(SHORT_DATE_FORMAT),
        'FID_INPUT_DATE_2': end_date.strftime(SHORT_DATE_FORMAT),
        'FID_PERIOD_DIV_CODE': 'D',
        'FID_ORG_ADJ_PRC': '0'
    }

    return get_api_response(url, headers, params)

# Search Functions
def get_search_conditions():
    """Retrieve search conditions from the API."""
    url = f"{settings['trading_url']}/uapi/domestic-stock/v1/quotations/psearch-title"
    headers = {
        'Content-Type': 'application/json; charset=utf-8',
        'Authorization': f'Bearer {get_valid_token()}',
        'appkey': settings['appkey'],
        'appsecret': settings['appsecret'],
        'tr_id': 'HHKST03900300'
    }
    params = {'user_id': settings['user_id']}
    logging.info(f"Request URL: {url}")
    logging.info(f"Request Headers: {headers}")
    logging.info(f"Request Params: {params}")

    search_conditions = []
    attempt_count = 0
    next_page_token = None

    while attempt_count < MAX_ATTEMPTS:
        if next_page_token:
            params['next_key'] = next_page_token

        response = get_api_response(url, headers, params)
        logging.info(f"API Response: {response}")

        if response and 'output2' in response:
            search_conditions.extend(response['output2'])
            break
        elif response and response.get('msg_cd') == 'MCA05762':
            logging.info("Handling pagination.")
            attempt_count += 1
            next_page_token = response.get('next_key')
            if not next_page_token:
                logging.error("Expected next_key for continuation but not found in response.")
                break
        else:
            logging.error(f"Unexpected response: {response}")
            break

    if attempt_count >= MAX_ATTEMPTS:
        logging.error("Reached maximum pagination attempts. Exiting.")
    logging.info(f"Search conditions: {search_conditions}")
    return search_conditions if search_conditions else None

def get_search_results(seq):
    """Retrieve search results from the API using the provided sequence."""
    url = f"{settings['trading_url']}/uapi/domestic-stock/v1/quotations/psearch-result"
    headers = {
        'Content-Type': 'application/json; charset=utf-8',
        'Authorization': f'Bearer {get_valid_token()}',
        'appkey': settings['appkey'],
        'appsecret': settings['appsecret'],
        'tr_id': 'HHKST03900400'
    }
    params = {'user_id': settings['user_id'], 'seq': seq}

    logging.info(f"Request URL: {url}")
    logging.info(f"Request Headers: {headers}")
    logging.info(f"Request Params: {params}")

    response = get_api_response(url, headers, params)
    logging.info(f"API Response: {response}")

    if isinstance(response, dict):
        if 'output2' in response:
            return response['output2']
        else:
            if response.get('msg_cd') == 'MCA05918':
                                logging.error(f"Condition search resulted in 0 stocks. Full response: {response}")
            else:
                logging.error(f"Missing 'output2' in response: {response}")
                logging.error(f"Full response: {response}")
            return None
    else:
        logging.error(f"Unexpected response format: {response}")
        return None

def save_data_to_csv(data, stock_code):
    """Save fetched stock data to a CSV file."""
    stock_code = stock_code.zfill(6)  # Ensure stock code is 6 characters long
    file_path = os.path.join(settings['data_path'], 'stock_data', f"{stock_code}_data.csv")
    os.makedirs(os.path.dirname(file_path), exist_ok=True)

    if 'output2' in data:
        output_data = []
        for item in data['output2']:
            # 조건 추가: date가 0이거나 거래량이 0이거나 음수인 경우 데이터를 무시합니다.
            if int(item.get('stck_bsop_date', '0')) == 0 or int(float(item.get('acml_tr_pbmn', '0').replace(',', '')) if item.get('acml_tr_pbmn') else 0) <= 0:
                continue

            output2_data = {
                'open': int(float(item.get('stck_oprc', '0').replace(',', '')) if item.get('stck_oprc') else 0),  # Opening price
                'high': int(float(item.get('stck_hgpr', '0').replace(',', '')) if item.get('stck_hgpr') else 0),  # High price
                'low': int(float(item.get('stck_lwpr', '0').replace(',', '')) if item.get('stck_lwpr') else 0),   # Low price
                'close': int(float(item.get('stck_clpr', '0').replace(',', '')) if item.get('stck_clpr') else 0),  # Closing price
                'volume': int(float(item.get('acml_tr_pbmn', '0').replace(',', '')) if item.get('acml_tr_pbmn') else 0),  # Accumulated trading volume
                'date': int(item.get('stck_bsop_date'))  # Business operation date
            }
            output_data.append(output2_data)
        
        if output_data:
            df = pd.DataFrame(output_data)
            df.to_csv(file_path, mode='a', index=False, header=not os.path.exists(file_path))
            logging.info(f"Stock data saved to {file_path}")
        else:
            logging.warning(f"No valid data to save for stock {stock_code}.")
    else:
        logging.error("Data format not recognized for saving to CSV.")

def load_last_saved_dates(file_path):
    """Load the last saved dates from the CSV file."""
    if os.path.exists(file_path):
        df = pd.read_csv(file_path)
        df['stock_code'] = df['stock_code'].apply(lambda x: str(x).zfill(6))  # Ensure stock_code is 6 digits
        return df.set_index('stock_code')['last_saved_date'].to_dict()
    else:
        return {}

def update_last_saved_dates(file_path, stock_code, last_saved_date):
    """Update the last saved dates in the CSV file."""
    df = pd.DataFrame([{'stock_code': stock_code.zfill(6), 'last_saved_date': last_saved_date}])
    
    if os.path.exists(file_path):
        existing_df = pd.read_csv(file_path)
        existing_df['stock_code'] = existing_df['stock_code'].apply(lambda x: str(x).zfill(6))
        combined_df = pd.concat([existing_df, df]).drop_duplicates(subset=['stock_code'], keep='last')
    else:
        combined_df = df

    combined_df.to_csv(file_path, index=False)

def get_stock_codes_and_names(seq):
    """Retrieve stock codes and names from search results."""
    search_conditions = get_search_conditions()
    if not search_conditions:
        logging.error("No search conditions found.")
        return []

    search_results = get_search_results(seq)
    if not search_results:
        logging.error("No search results found.")
        return []

    stock_codes_and_names = []
    for result in search_results:
        code = result.get('code')
        name = result.get('name')
        logging.info(f"Retrieved code: {code}, name: {name}")
        if code and name:
            stock_codes_and_names.append((code.zfill(6), name))
        else:
            logging.error(f"Invalid stock code or name: {result}")

    logging.info(f"Detected stock codes and names: {stock_codes_and_names}")
    save_stock_codes_to_csv(stock_codes_and_names)
    return stock_codes_and_names

def save_stock_codes_to_csv(stock_codes_and_names):
    """Save stock codes and names to a CSV file."""
    file_path = os.path.join(settings['data_path'], 'stock_codes.csv')
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    
    if os.path.exists(file_path):
        existing_df = pd.read_csv(file_path)
    else:
        existing_df = pd.DataFrame(columns=['code', 'name'])
    
    stock_codes_and_names = [(code.zfill(6), name) for code, name in stock_codes_and_names]
    new_df = pd.DataFrame(stock_codes_and_names, columns=['code', 'name'])
    combined_df = pd.concat([existing_df, new_df]).drop_duplicates().reset_index(drop=True)
    
    combined_df.to_csv(file_path, index=False)
    logging.info(f"Stock codes saved to {file_path}")

def fetch_and_save_data_with_dates(stock_codes_and_names, last_saved_dates):
    """Fetch and save stock data, updating last saved dates."""
    today = datetime.now().date()

    for stock_code, stock_name in stock_codes_and_names:
        stock_code = stock_code.zfill(6)  # Ensure stock code is 6 digits
        last_saved_date = last_saved_dates.get(stock_code)
        start_date = datetime.strptime(last_saved_date, "%Y-%m-%d").date() if last_saved_date else today - timedelta(days=365 * 10)
        
        if start_date < today:
            # Fetch data from the last saved date to today
            data = fetch_stock_data(stock_code, start_date, today)
            if data:
                save_data_to_csv(data, stock_code)
                update_last_saved_dates('path_to_last_saved_dates.csv', stock_code, today.strftime("%Y-%m-%d"))
            else:
                logging.error(f"No data fetched for stock: {stock_code}")
        else:
            logging.info(f"Data for stock {stock_code} is already up to date.")

def fetch_data_for_detected_stocks():
    """Main function to fetch data for detected stocks."""
    last_saved_dates = load_last_saved_dates('path_to_last_saved_dates.csv')
    
    for seq in range(2):
        detected_stocks = get_stock_codes_and_names(seq)
        if detected_stocks:
            fetch_and_save_data_with_dates(detected_stocks, last_saved_dates)
            break
        else:
            logging.info(f"No stocks found for seq: {seq}. Trying next seq.")

def main():
    try:
        fetch_data_for_detected_stocks()
    except Exception as e:
        logging.error(f"An error occurred: {e}")

if __name__ == "__main__":
    main()
