import os
import pandas as pd
from sklearn.preprocessing import StandardScaler
import logging
from sklearn.model_selection import train_test_split
from feature_engineering import add_features
import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'scripts', 'model')))

from scripts.model.model_training import update_lstm_model, train_lstm_model

logging.basicConfig(level=logging.INFO)

def normalize_data(df):
    """Normalize the data using StandardScaler."""
    scaler = StandardScaler()
    normalized_df = pd.DataFrame(scaler.fit_transform(df), columns=df.columns)
    return normalized_df, scaler

def load_csv_files_from_directory(directory_path):
    dataframes = []
    for file_name in os.listdir(directory_path):
        if file_name.endswith('.csv'):
            file_path = os.path.join(directory_path, file_name)
            try:
                df = pd.read_csv(file_path, encoding='utf-8')
                logging.info(f"Loaded {file_name} with columns: {df.columns}")
                df.columns = df.columns.str.lower()  # Make all column names lowercase
                if not df.empty:
                    logging.info(f"First few rows of {file_name}:\n{df.head()}")
                    # Check for required columns based on the type of data
                    if 'news_data' in directory_path:
                        required_columns = ['title', 'link', 'date', 'stock_name', 'stock_code']
                    elif 'financial_data' in directory_path:
                        required_columns = ['date', 'account_nm', 'thstrm_amount', 'frmtrm_amount']
                    elif 'economic_data' in directory_path and 'gdp_data' in file_name:
                        required_columns = ['date', 'item code', 'item name', 'gdp']
                    elif 'economic_data' in directory_path and 'inflation_rate_data' in file_name:
                        required_columns = ['date', 'item code', 'item name', 'inflation rate']
                    else:
                        required_columns = ['open', 'high', 'low', 'close', 'volume', 'date']
                    
                    if all(col in df.columns for col in required_columns):
                        dataframes.append(df)
                    else:
                        missing_cols = set(required_columns) - set(df.columns)
                        logging.error(f"Missing required columns {missing_cols} in {file_path}")
                else:
                    logging.warning(f"Empty file: {file_path}")
            except pd.errors.EmptyDataError:
                logging.warning(f"Empty file: {file_path}")
            except Exception as e:
                logging.error(f"Error reading {file_path}: {e}")
    return pd.concat(dataframes, ignore_index=True) if dataframes else pd.DataFrame()

def preprocess_df(df, date_column='date', date_format='%Y%m%d'):
    if date_column not in df.columns:
        raise ValueError(f"Date column '{date_column}' not found in the dataset. Available columns: {df.columns}")
    df[date_column] = pd.to_datetime(df[date_column], format=date_format, errors='coerce')
    df.dropna(subset=[date_column], inplace=True)
    df.fillna(0, inplace=True)
    return df

def preprocess_data(stock_data, financial_data, news_data, economic_data):
    stock_data = preprocess_df(stock_data, date_column='date', date_format='%Y%m%d')
    financial_data = preprocess_df(financial_data, date_column='date', date_format='%Y')
    news_data = preprocess_df(news_data, date_column='date', date_format='%Y%m%d')
    economic_data = preprocess_df(economic_data, date_column='date', date_format='%Y-%m-%d')

    stock_data['target'] = stock_data['close'].shift(-1)
    stock_data.dropna(subset=['target'], inplace=True)

    logging.info(f"Stock data after preprocessing: {stock_data.head()}")
    logging.info(f"Financial data after preprocessing: {financial_data.head()}")
    logging.info(f"Economic data after preprocessing: {economic_data.head()}")
    logging.info(f"News data after preprocessing: {news_data.head()}")

    # Check if financial_data or news_data is empty and skip merging if they are
    if not financial_data.empty:
        stock_data = pd.merge_asof(stock_data.sort_values('date'), 
                                   financial_data.sort_values('date'), 
                                   on='date', direction='nearest', suffixes=('', '_financial'))
    if not economic_data.empty:
        stock_data = pd.merge_asof(stock_data.sort_values('date'), 
                                   economic_data.sort_values('date'), 
                                   on='date', direction='nearest', suffixes=('', '_economic'))
    if not news_data.empty:
        stock_data = pd.merge_asof(stock_data.sort_values('date'), 
                                   news_data.sort_values('date'), 
                                   on='date', direction='nearest', suffixes=('', '_news'))

    logging.info(f"Combined data after merging: {stock_data.head()}")

    stock_data = add_features(stock_data)

    for col in stock_data.columns:
        if stock_data[col].dtype == 'object':
            stock_data[col] = pd.to_numeric(stock_data[col].str.replace(',', ''), errors='coerce')

    stock_data.ffill(inplace=True)
    stock_data.bfill(inplace=True)
    
    # Handle remaining NaN values by filling them with 0
    stock_data.fillna(0, inplace=True)

    X = stock_data.drop(['target', 'date'], axis=1)
    y = stock_data['target']

    if X.empty or y.empty:
        logging.error("Insufficient data after preprocessing for training the model.")
        return None, None, None

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)
    scaler = StandardScaler()
    X_train = scaler.fit_transform(X_train)
    X_test = scaler.transform(X_test)

    return (X_train, X_test, y_train, y_test), scaler, stock_data

def main(base_dir):
    stock_data_dir = os.path.join(base_dir, 'stock_data')
    financial_data_dir = os.path.join(base_dir, 'financial_data')
    news_data_dir = os.path.join(base_dir, 'news_data')
    economic_data_dir = os.path.join(base_dir, 'economic_data')

    stock_data = load_csv_files_from_directory(stock_data_dir)
    financial_data = load_csv_files_from_directory(financial_data_dir)
    news_data = load_csv_files_from_directory(news_data_dir)
    economic_data = load_csv_files_from_directory(economic_data_dir)

    if stock_data.empty:
        logging.error("Stock data is missing or empty.")
        return
    if financial_data.empty:
        logging.error("Financial data is missing or empty.")
    if news_data.empty:
        logging.error("News data is missing or empty.")
    if economic_data.empty:
        logging.error("Economic data is missing or empty.")
    
    result = preprocess_data(stock_data, financial_data, news_data, economic_data)
    
    if result is None:
        logging.error("Data preprocessing failed due to insufficient data.")
        return

    (X_train, X_test, y_train, y_test), scaler, combined_data = result

    if combined_data is None:
        logging.error("Data preprocessing failed.")
        return

    output_path = os.path.join(base_dir, 'preprocessed_data.csv')
    combined_data.to_csv(output_path, index=False)
    logging.info(f"Combined data saved to {output_path}")

if __name__ == "__main__":
    base_dir = 'C:\\Users\\user\\projects\\st\\data'  # 사용자의 데이터 디렉토리 경로
    main(base_dir)
