import os
import pandas as pd
from textblob import TextBlob
import logging
from settings import settings
from data_collection import read_stock_data, check_and_fix_columns
from model_training import train_lstm_model, update_lstm_model, save_model, load_model
from trading_logic import generate_trading_signals, execute_trades

class NewsSentimentAnalyzer:
    def __init__(self, news_data_path):
        self.news_data_path = news_data_path

    def fetch_news(self, stock_code):
        news_data_file = os.path.join(self.news_data_path, f"news_data_{stock_code}.csv")
        if os.path.exists(news_data_file):
            return pd.read_csv(news_data_file)
        else:
            logging.error(f"News data file for {stock_code} not found at {news_data_file}")
            return pd.DataFrame()

    def analyze_sentiment(self, news_data):
        if news_data.empty:
            return 0
        sentiments = []
        for _, article in news_data.iterrows():
            text = f"{article['title']}. {article['description']}"
            sentiment = TextBlob(text).sentiment.polarity
            sentiments.append(sentiment)
        average_sentiment = sum(sentiments) / len(sentiments) if sentiments else 0
        return average_sentiment

    def get_stock_sentiment(self, stock_code):
        news_data = self.fetch_news(stock_code)
        sentiment = self.analyze_sentiment(news_data)
        return sentiment

def integrate_sentiment_with_model(stock_code):
    sentiment_analyzer = NewsSentimentAnalyzer(news_data_path=os.path.join(settings['data_path'], 'news_data'))
    sentiment = sentiment_analyzer.get_stock_sentiment(stock_code)
    
    # Load stock data
    stock_data_path = os.path.join(settings['data_path'], f"{stock_code}_data.csv")
    if not os.path.exists(stock_data_path):
        logging.error(f"Stock data file for {stock_code} not found at {stock_data_path}")
        return

    stock_data = pd.read_csv(stock_data_path)
    if stock_data is None or stock_data.empty:
        logging.error(f"No stock data found for {stock_code}")
        return

    stock_data = check_and_fix_columns(stock_data)

    # Integrate sentiment into stock data
    stock_data['sentiment'] = sentiment

    # Train or update LSTM model
    lstm_model_path = os.path.join(settings['model_path'], f"lstm_{stock_code}.h5")
    if os.path.exists(lstm_model_path):
        model = load_model(lstm_model_path)
        model = update_lstm_model(stock_data[['Close']], model)
    else:
        model = train_lstm_model(stock_data[['Close']])
    save_model(model, lstm_model_path)

    # Generate trading signals based on combined data
    stock_data = generate_trading_signals(stock_data)
    
    # Execute trades
    execute_trades(stock_data)

def main():
    stock_codes = get_stock_codes()  # 주식 코드 목록 가져오기
    for stock_code in stock_codes:
        try:
            integrate_sentiment_with_model(stock_code)
        except Exception as e:
            logging.error(f"Error processing stock {stock_code}: {e}")

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    main()
